/* Part of the SUDK sudoku puzzle solver/editor */
/* By Alex Wilson, copyright 2006 */

/* Solving algorithm based in part on the freeware work of
   Wim Colgate */

#include "sudoku.h"

int sudoku_get(sudoku *board, int index) {
	if (index % 2 == 0) {
		return (board->harray[index >> 1] & 240) >> 4;
	} else {
		return board->harray[index >> 1] & 15;
	}
}

void sudoku_set(sudoku *board, int index, int value) {
	if (index % 2 == 0) {
		board->harray[index >> 1] = ((value << 4) & 240) | (board->harray[index >> 1] & 15);
	} else {
		board->harray[index >> 1] = (value & 15) | (board->harray[index >> 1] & 240);
	}
}

void drawOutline() {
	int y, x;

	for (y=0; y <= LCD_HEIGHT-4; y += CELL_HEIGHT) {
		DrawLine(0, y, CELL_WIDTH*9, y, A_NORMAL);
	}

	for (y=0; y <= LCD_HEIGHT-4; y += CELL_HEIGHT*3) {
		DrawLine(0, y, CELL_WIDTH*9, y, A_THICK1);
	}

	for (x=0; x <= LCD_WIDTH; x += CELL_WIDTH) {
		DrawLine(x, 0, x, CELL_HEIGHT*9, A_NORMAL);
	}

	for (x=0; x <= LCD_WIDTH; x += CELL_WIDTH*3) {
		DrawLine(x, 0, x, CELL_HEIGHT*9, A_THICK1);
	}
}

void highlight(int id) {
	int row = (id - (id % 9)) / 9;
	int col = (id % 9);

	ScrRectFill(
		&(SCR_RECT) { CELL_WIDTH * col, CELL_HEIGHT * row, CELL_WIDTH * (col + 1), CELL_HEIGHT * (row + 1) },
		&(SCR_RECT) { 0,0,LCD_WIDTH, LCD_HEIGHT},
	A_XOR);
}

void highlight_row(int row) {
	long int i;

	ScrRectFill(
		&(SCR_RECT) { 0, CELL_HEIGHT * row, CELL_WIDTH * 9, CELL_HEIGHT * (row+1) },
		&(SCR_RECT) { 0,0,LCD_WIDTH,LCD_HEIGHT },
	A_XOR);

	for (i=0; i<=600000; i++) { }

	ScrRectFill(
		&(SCR_RECT) { 0, CELL_HEIGHT * row, CELL_WIDTH * 9, CELL_HEIGHT * (row+1) },
		&(SCR_RECT) { 0,0,LCD_WIDTH,LCD_HEIGHT },
	A_XOR);
}

void highlight_column(int col) {
	long int i;

	ScrRectFill(
		&(SCR_RECT) { CELL_WIDTH * col, 0, CELL_WIDTH * (col+1), CELL_HEIGHT*9 },
		&(SCR_RECT) {0,0,LCD_WIDTH,LCD_HEIGHT},
	A_XOR);

	for (i=0; i<=600000; i++) { }

	ScrRectFill(
		&(SCR_RECT) { CELL_WIDTH * col, 0, CELL_WIDTH * (col+1), CELL_HEIGHT*9 },
		&(SCR_RECT) {0,0,LCD_WIDTH,LCD_HEIGHT},
	A_XOR);
}

void highlight_square(int basecol, int baserow) {
	long int i;

	ScrRectFill(
		&(SCR_RECT) { CELL_WIDTH * basecol, CELL_HEIGHT * baserow, CELL_WIDTH * (basecol+3), CELL_HEIGHT * (baserow+3)},
		&(SCR_RECT) {0,0,LCD_WIDTH,LCD_HEIGHT},
	A_XOR);

	for (i=0; i<=600000; i++) { }

	ScrRectFill(
		&(SCR_RECT) { CELL_WIDTH * basecol, CELL_HEIGHT * baserow, CELL_WIDTH * (basecol+3), CELL_HEIGHT * (baserow+3)},
		&(SCR_RECT) {0,0,LCD_WIDTH,LCD_HEIGHT},
	A_XOR);
}

void drawBoard(sudoku *board) {
	int row, cell;

	FontSetSys(F_4x6);

	for (row=0; row < 81; row += 9) {
		for (cell=row; cell <= row+8; cell++) {
			if (sudoku_get(board,cell) > 0)
				DrawChar(CELL_WIDTH*(cell-row) + 3, CELL_HEIGHT*(row/9) + 3, '0' + sudoku_get(board,cell), A_XOR);
		}
	}
}

void drawCell(sudoku *board, int cell) {
	int row = (cell - (cell % 9));

	FontSetSys(F_4x6);

	ScrRectFill(
		&(SCR_RECT) { CELL_WIDTH*(cell-row) + 3, CELL_HEIGHT*(row/9) + 3, CELL_WIDTH*(cell-row) + 10, CELL_HEIGHT*(row/9) + 8 },
		&(SCR_RECT) { 0,0,LCD_WIDTH, LCD_HEIGHT},
	A_REVERSE);

	if (sudoku_get(board,cell) > 0)
		DrawChar(CELL_WIDTH*(cell-row) + 3, CELL_HEIGHT*(row/9) + 3, '0' + sudoku_get(board,cell), A_XOR);
}

result check(sudoku *board) {
	ST_helpMsg("Please wait...");

	result t,r;
	t = checkRows(board);
	r.failures = t.failures;
	r.blanks = t.blanks;

	t = checkColumns(board);
	r.failures += t.failures;

	t = checkBoxes(board);
	r.failures += t.failures;

	return r;
}

result checkBoxes(sudoku *board) {
	int bigrow, bigcol, row, cell, i;
	int check[10];
	result temp;
	temp.failures = 0; temp.blanks = 0;

	for (bigrow = 0; bigrow < 81; bigrow += 27) {
		for (bigcol = bigrow; bigcol < bigrow+9; bigcol += 3) {
			for (i=0; i<10; i++) check[i] = 0;
			for (row = bigcol; row < bigcol + 27; row += 9) {
				for (cell = row; cell < row + 3; cell++) {
					check[sudoku_get(board,cell)]++;
				}
			}

			for (i=1; i<10; i++) {
				if (check[i] > 1) {
					highlight_square(bigcol-bigrow, bigrow/9);
					temp.failures++;
				}
			}
			if (check[0] > 0) temp.blanks+=check[0];
		}
	}

	return temp;
}

result checkRows(sudoku *board) {
	int row, cell, i;
	int check[10];
	result temp; temp.failures = 0; temp.blanks = 0;

	for (row=0; row < 81; row += 9) {
		for (i=0; i<10; i++) check[i] = 0;
		for (cell=row; cell <= row + 8; cell++) {
			check[sudoku_get(board,cell)]++;
		}

		for (i=1; i<10; i++) {
			if (check[i] > 1) {
				highlight_row(row/9);
				temp.failures++;
			}
		}
		if (check[0] > 0) temp.blanks+=check[0];
	}

	return temp;
}

result checkColumns(sudoku *board) {
	int column, cell, i;
	int check[10];
	result temp; temp.failures = 0; temp.blanks = 0;

	for (column=0; column <= 8; column++) {
		for (i=0; i<10; i++) check[i] = 0;
		for (cell=column; cell < 81; cell += 9) {
			check[sudoku_get(board,cell)]++;
		}

		for (i=1; i<10; i++) {
			if (check[i] > 1) {
				highlight_column(column);
				temp.failures++;
			}
		}
		if (check[0] > 0) temp.blanks+=check[0];
	}

	return temp;
}
